/* -*- Mode: IDL; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Tim Miller <mille449@gmail.com>.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Siddharth Agarwal <sid.bugzilla@gmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsISupports.idl"

interface imgIContainer;

/**
 * Starting in Windows 7, applications can display an overlay on the icon in
 * the taskbar. This class wraps around the native functionality to do this.
 */
[scriptable, uuid(b1858889-a698-428a-a14b-b5d60cff6de2)]
interface nsITaskbarOverlayIconController : nsISupports
{
  /**
   * Sets the overlay icon and its corresponding alt text.
   *
   * @param statusIcon The handle to the overlay icon. The icon will be scaled
   *                   to the small icon size (16x16 at 96 dpi). Can be null, in
   *                   which case if the taskbar button represents a single window
   *                   the icon is removed.
   * @param statusDescription The alt text version of the information
   *                          conveyed by the overlay, for accessibility
   *                          purposes.
   *
   * @note The behavior for window groups is managed by Windows.
   * - If an overlay icon is set for any window in a window group and another
   *   overlay icon is already applied to the corresponding taskbar button, that
   *   existing overlay is replaced.
   * - If null is passed in to replace the overlay currently being displayed,
   *   and if a previous overlay set for a different window in the group is
   *   still available, then that previous overlay is displayed.
   */
  void setOverlayIcon(in imgIContainer statusIcon,
                      in AString statusDescription);
};
